#include <stdio.h>
#include <nicklib.h>
#include <stdlib.h>
#include "eigsubs.h"
#include <unistd.h>

#define N 496 /* size of matrix */
#define P 2432 /* number of pairs */
#define MAXN 1500 /* max sample number */
#define M 18735 /* number of genes = 18735 */
#define ITER 5
#define NROUND 5
#define VERBOSE 0
#define INFILE "/groups/price/aprice/DECODEGE2/PHASED/FAMGEABIBD.782.norm"
#define PAIRSFILE "/groups/price/aprice/DECODEGE2/PHASED/FAMALLIBD"  
/* N.B. PAIRSFILE also includes theta information */

double *dataIFB, *dataIFA, *bigdata;
double *eval, *evec, *bigeval, *bigevec;
double h2IFB, h2cisIFB, h2IFA, h2cisIFA, rho, rhocorr, xi2, xi2cis;
double *theta, *gammax;
double *coeff, *cinv, *bigcoeff, *bigcinv, temp;
double compute_logl_IFB();
double compute_logl_IFA();
double compute_logl();

int main(int argc, char **argv)
{
  double step, bestx, bestxcis, logl, bestlogl, *ptheta;
  double sumx, sumxx, sumy, sumyy, sumxy, sum1;
  double meanx, meany, sdevx, sdevy;
  int k, n, n2, m, tempn, i, iter, p, N1, N2, round;
  int *goodp, ngoodp, *nN, *Nn, *pn1, *pn2, mstart, mend;
  char tempstring[3000], Xchar;
  char *OUTFILE = NULL;
  FILE *fpin, *fppairs, *fpout;

  while((i = getopt(argc,argv,"s:e:o:")) != -1)
    {
      switch(i)
        {
        case 's':
          mstart = atoi(optarg);
          break;
        case 'e':
          mend = atof(optarg);
          break;
        case 'o':          /* input genotype file */
          OUTFILE = (char *) strdup(optarg);
          break;
        }
    }

  if((coeff = (double *) malloc(N*N*sizeof(*coeff))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((cinv = (double *) malloc(N*N*sizeof(*cinv))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((eval = (double *) malloc(N*sizeof(*eval))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((evec = (double *) malloc(N*N*sizeof(*evec))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((bigcoeff = (double *) malloc(2*N*2*N*sizeof(*bigcoeff))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((bigcinv = (double *) malloc(2*N*2*N*sizeof(*bigcinv))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((bigeval = (double *) malloc(2*N*sizeof(*bigeval))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((bigevec = (double *) malloc(2*N*2*N*sizeof(*bigevec))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((theta = (double *) malloc(N*N*sizeof(*theta))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((gammax = (double *) malloc(N*N*sizeof(*gammax))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((dataIFB = (double *) malloc(N*sizeof(*dataIFB))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((dataIFA = (double *) malloc(N*sizeof(*dataIFA))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((bigdata = (double *) malloc(2*N*sizeof(*bigdata))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((nN = (int *) malloc(N*sizeof(*nN))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((Nn = (int *) malloc(MAXN*sizeof(*Nn))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((goodp = (int *) malloc(P*sizeof(*goodp))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((pn1 = (int *) malloc(P*sizeof(*pn1))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((pn2 = (int *) malloc(P*sizeof(*pn2))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((ptheta = (double *) malloc(P*sizeof(*ptheta))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  
  if( (fpin = fopen(INFILE, "r")) == NULL)
    {
      fprintf(stderr,"Could not open input file %s\n", INFILE);  exit(1);
    }
  if( (fppairs = fopen(PAIRSFILE, "r")) == NULL)
    {
      fprintf(stderr,"Could not open input file %s\n", PAIRSFILE);  exit(1);
    }
  if( (fpout = fopen(OUTFILE, "w")) == NULL)
    {
      fprintf(stderr,"Could not open output file %s\n", OUTFILE);  exit(1);
    }

  /* PROCESS 1ST LINE OF DATA FILE TO MAKE nN[] and Nn[] */
  fscanf(fpin,"%s",&tempstring); 
  for(k=0; k<MAXN; k++) Nn[k] = -1; /* -1 means absent */
  for(n=0; n<N; n++)
  {
    fscanf(fpin,"%d",&nN[n]);
    Nn[nN[n]] = n;
  }
  /* use PAIRSFILE to make pn1[p] and pn2[p] and ptheta[] */
  ngoodp = 0;
  for(p=0; p<P; p++)
  {
    goodp[p] = 0;
    fscanf(fppairs,"%lf",&ptheta[p]);
    fscanf(fppairs,"%d",&N1);
    fscanf(fppairs,"%d",&N2);
    if(Nn[N1]<0) continue; if(Nn[N2]<0) continue;
    goodp[p] = 1; ngoodp++;
    pn1[p] = Nn[N1]; pn2[p] = Nn[N2]; 
  }
  /* use ptheta[p] to make theta matrix */
  for(p=0; p<P; p++)
  {
    if(goodp[p] == 0) continue;
    theta[pn1[p]+N*pn2[p]] = ptheta[p];
    theta[pn2[p]+N*pn1[p]] = ptheta[p];
  }
  /* PROCESS GENE EXPRESSION DATA */
  for(m=0; m<M; m++)
  {
    /* make dataIFA[] and dataIFB[] and gammax[] */
    fscanf(fpin,"%s",&tempstring);
    for(n=0; n<N; n++) 
    {
      fscanf(fpin,"%lf",&dataIFA[n]); bigdata[n+N] = dataIFA[n];
      fscanf(fpin,"%lf",&dataIFB[n]); bigdata[n] = dataIFB[n];
    }
    fscanf(fpin,"%c",&Xchar); /* space character */
    for(n=0; n<N; n++)
    { 
      for(n2=0; n2<N; n2++) gammax[n+N*n2] = 0.0;
    }
    for(p=0; p<P; p++)
    {
      fscanf(fpin,"%c",&Xchar);
      if(goodp[p] == 0) continue;
      if(Xchar == '1') 
        { gammax[pn1[p]+N*pn2[p]] = 0.5; gammax[pn2[p]+N*pn1[p]] = 0.5; }
      if(Xchar == '2') 
        { gammax[pn1[p]+N*pn2[p]] = 1.0; gammax[pn2[p]+N*pn1[p]] = 1.0; }
    }
    if(m < mstart) continue;
    if(m >= mend) break;

    /* ESTIMATE h2IFB: binary search */
    h2IFB = 0.50; h2cisIFB = 0.00; logl = compute_logl_IFB();
    bestx = h2IFB; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2IFB = bestx-step; logl = compute_logl_IFB();
      if(logl>bestlogl) { bestx=h2IFB; bestlogl=logl; step /= 2.0; continue; }
      h2IFB = bestx+step; logl = compute_logl_IFB();
      if(logl>bestlogl) { bestx=h2IFB; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2IFB = bestx;

    /* ESTIMATE h2cisIFB: binary search */
    h2cisIFB = 0.50; logl = compute_logl_IFB();
    bestx = h2cisIFB; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2cisIFB = bestx-step; logl = compute_logl_IFB();
      if(logl > bestlogl) 
        { bestx=h2cisIFB; bestlogl=logl; step /= 2.0; continue; }
      h2cisIFB = bestx+step; logl = compute_logl_IFB();
      if(logl > bestlogl) 
        { bestx=h2cisIFB; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2cisIFB = bestx;

    /* ESTIMATE h2IFA: binary search */
    h2IFA = 0.50; h2cisIFA = 0.00; logl = compute_logl_IFA();
    bestx = h2IFA; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2IFA = bestx-step; logl = compute_logl_IFA();
      if(logl>bestlogl) { bestx=h2IFA; bestlogl=logl; step /= 2.0; continue; }
      h2IFA = bestx+step; logl = compute_logl_IFA();
      if(logl>bestlogl) { bestx=h2IFA; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2IFA = bestx;

    /* ESTIMATE h2cisIFA: binary search */
    h2cisIFA = 0.50; logl = compute_logl_IFA();
    bestx = h2cisIFA; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2cisIFA = bestx-step; logl = compute_logl_IFA();
      if(logl > bestlogl) 
        { bestx=h2cisIFA; bestlogl=logl; step /= 2.0; continue; }
      h2cisIFA = bestx+step; logl = compute_logl_IFA();
      if(logl > bestlogl) 
        { bestx=h2cisIFA; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2cisIFA = bestx;

    /* ESTIMATE rho: compute correlation */
    sumx=0.0; sumxx=0.0; sumy=0.0; sumyy=0.0; sumxy=0.0; sum1=0.0;
    for(n=0; n<N; n++)
    {
      sumx += dataIFB[n]; sumxx += dataIFB[n]*dataIFB[n];
      sumy += dataIFA[n]; sumyy += dataIFA[n]*dataIFA[n];
      sum1 += 1.0; sumxy += dataIFB[n]*dataIFA[n];
    }
    meanx = sumx/sum1; meany = sumy/sum1;
    sdevx = sqrt(sumxx/sum1 - meanx*meanx);
    sdevy = sqrt(sumyy/sum1 - meany*meany);
    rhocorr = (sumxy/sum1 - meanx*meany) / (sdevx*sdevy);
    if(VERBOSE) fprintf(stderr,"correlation rho = %.03f\n",rhocorr); 

    /* ESTIMATE rho: binary search */
    rho = 0.0; xi2 = 0.0; xi2cis = 0.0; logl = compute_logl();
    bestx = rho; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      rho = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=rho; bestlogl=logl; step /= 2.0; continue; }
      rho = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=rho; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    rho = bestx;
    if(VERBOSE) fprintf(stderr,"likelihood rho = %.03f\n",rho); 

    /* ESTIMATE xi2: binary search */
    xi2 = 0.50; logl = compute_logl();
    bestx = xi2; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      xi2 = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2; bestlogl=logl; step /= 2.0; continue; }
      xi2 = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    xi2 = bestx;

    /* ESTIMATE xi2cis: binary search */
    xi2cis = 0.50; logl = compute_logl();
    bestx = xi2cis; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      xi2cis = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2cis; bestlogl=logl; step /= 2.0; continue; }
      xi2cis = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2cis; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    xi2cis = bestx;

    /* ROUND r */
    for(round=1; round<NROUND; round++)
    {
    /* ESTIMATE h2IFB: binary search */
    /* h2IFB = 0.50; h2cisIFB = 0.00; logl = compute_logl_IFB(); */
    bestx = h2IFB; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2IFB = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=h2IFB; bestlogl=logl; step /= 2.0; continue; }
      h2IFB = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=h2IFB; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2IFB = bestx;

    /* ESTIMATE h2cisIFB: binary search */
    /* h2cisIFB = 0.50; logl = compute_logl_IFB(); */
    bestx = h2cisIFB; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2cisIFB = bestx-step; logl = compute_logl();
      if(logl > bestlogl) 
        { bestx=h2cisIFB; bestlogl=logl; step /= 2.0; continue; }
      h2cisIFB = bestx+step; logl = compute_logl();
      if(logl > bestlogl) 
        { bestx=h2cisIFB; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2cisIFB = bestx;

    /* ESTIMATE h2IFA: binary search */
    /* h2IFA = 0.50; h2cisIFA = 0.00; logl = compute_logl_IFA(); */
    bestx = h2IFA; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2IFA = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=h2IFA; bestlogl=logl; step /= 2.0; continue; }
      h2IFA = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=h2IFA; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2IFA = bestx;

    /* ESTIMATE h2cisIFA: binary search */
    /* h2cisIFA = 0.50; logl = compute_logl_IFA(); */
    bestx = h2cisIFA; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2cisIFA = bestx-step; logl = compute_logl();
      if(logl > bestlogl) 
        { bestx=h2cisIFA; bestlogl=logl; step /= 2.0; continue; }
      h2cisIFA = bestx+step; logl = compute_logl();
      if(logl > bestlogl) 
        { bestx=h2cisIFA; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2cisIFA = bestx;

    /* ESTIMATE rho: binary search */
    /* rho = 0.0; xi2 = 0.0; xi2cis = 0.0; logl = compute_logl(); */
    bestx = rho; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      rho = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=rho; bestlogl=logl; step /= 2.0; continue; }
      rho = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=rho; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    rho = bestx;
    if(VERBOSE) fprintf(stderr,"likelihood rho = %.03f\n",rho); 

    /* ESTIMATE xi2: binary search */
    /* xi2 = 0.50; logl = compute_logl(); */
    bestx = xi2; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      xi2 = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2; bestlogl=logl; step /= 2.0; continue; }
      xi2 = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    xi2 = bestx;

    /* ESTIMATE xi2cis: binary search */
    /* xi2cis = 0.50; logl = compute_logl(); */
    bestx = xi2cis; logl = bestlogl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      xi2cis = bestx-step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2cis; bestlogl=logl; step /= 2.0; continue; }
      xi2cis = bestx+step; logl = compute_logl();
      if(logl>bestlogl) { bestx=xi2cis; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    xi2cis = bestx;
    }

    fprintf(fpout,"%.03f %.03f %.03f   %.03f %.03f %.03f   %.03f   %.03f %.03f %.03f => %.03f\n",h2IFB,h2cisIFB,h2IFB-h2cisIFB,h2IFA,h2cisIFA,h2IFA-h2cisIFA,rhocorr,rho,xi2,xi2cis,bestlogl);
  }
}

double compute_logl_IFB()
{
  int k, n, n2;
  double det, logl;

  /* spectral decomposition */
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<N; n2++) 
      coeff[n+N*n2] = h2cisIFB*gammax[n+N*n2] + (h2IFB-h2cisIFB)*theta[n+N*n2];
    coeff[n+N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  pdinv(cinv, coeff, N); 
  eigvecs(coeff, eval, evec, N);

  det = 1.0;
  for(k=0; k<N; k++) det *= eval[k];
  if(det < 0.0) return -1000000000.0;
  logl = -0.5*log(det);
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<n; n2++) /* n2<n due to symmetry */
      logl -= dataIFB[n]*dataIFB[n2]*cinv[n+N*n2]; /* 2.0 * 0.5 = 1.0 */
    n2 = n;
      logl -= 0.5*dataIFB[n]*dataIFB[n2]*cinv[n+N*n2];
  }
  if(VERBOSE) fprintf(stderr,"h2IFB=%.03f h2cisIFB=%.03f logl=%.03f\n",h2IFB,h2cisIFB,logl); 
  if(!((logl > -1000000000.0) && (logl < 1000000000.0))) /* nan */
    return -1000000000.0;
  return logl;
}

double compute_logl_IFA()
{
  int k, n, n2;
  double det, logl;

  /* spectral decomposition */
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<N; n2++) 
      coeff[n+N*n2] = h2cisIFA*gammax[n+N*n2] + (h2IFA-h2cisIFA)*theta[n+N*n2];
    coeff[n+N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  pdinv(cinv, coeff, N); 
  eigvecs(coeff, eval, evec, N);

  det = 1.0;
  for(k=0; k<N; k++) det *= eval[k];
  if(det < 0.0) return -1000000000.0;
  logl = -0.5*log(det);
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<n; n2++) /* n2<n due to symmetry */
      logl -= dataIFA[n]*dataIFA[n2]*cinv[n+N*n2]; /* 2.0 * 0.5 = 1.0 */
    n2 = n;
      logl -= 0.5*dataIFA[n]*dataIFA[n2]*cinv[n+N*n2];
  }
  if(VERBOSE) fprintf(stderr,"h2IFA=%.03f h2cisIFA=%.03f logl=%.03f\n",h2IFA,h2cisIFA,logl); 
  if(!((logl > -1000000000.0) && (logl < 1000000000.0))) /* nan */
    return -1000000000.0;
  return logl;
}

double compute_logl()
{
  int k, n, n2, a, a2;
  double det, logl;

  /* spectral decomposition */
  for(a=0; a<N; a++)
  {
    n = a;
    for(a2=0; a2<N; a2++) 
    {
      n2 = a2;
      bigcoeff[n+2*N*n2] = h2cisIFB*gammax[a+N*a2] + (h2IFB-h2cisIFB)*theta[a+N*a2];
      /* abuse of notation: in this program, theta=gamma=0 on diagonal */
    }
    bigcoeff[n+2*N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  for(a=0; a<N; a++)
  {
    n = a+N;
    for(a2=0; a2<N; a2++) 
    {
      n2 = a2+N;
      bigcoeff[n+2*N*n2] = h2cisIFA*gammax[a+N*a2] + (h2IFA-h2cisIFA)*theta[a+N*a2];
      /* abuse of notation: in this program, theta=gamma=0 on diagonal */
    }
    bigcoeff[n+2*N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  for(a=0; a<N; a++)
  {
    n = a;
    for(a2=0; a2<N; a2++) 
    {
      n2 = a2+N;
      bigcoeff[n+2*N*n2] = xi2cis*gammax[a+N*a2] + (xi2-xi2cis)*theta[a+N*a2];
      /* abuse of notation: in this program, theta=gamma=0 on diagonal */
    }
    bigcoeff[a+2*N*(a+N)] += rho; 
  }
  for(a=0; a<N; a++)
  {
    n = a+N;
    for(a2=0; a2<N; a2++) 
    {
      n2 = a2;
      bigcoeff[n+2*N*n2] = xi2cis*gammax[a+N*a2] + (xi2-xi2cis)*theta[a+N*a2];
      /* abuse of notation: in this program, theta=gamma=0 on diagonal */
    }
    bigcoeff[(a+N)+2*N*a] += rho; 
  }
  pdinv(bigcinv, bigcoeff, 2*N); 
  eigvecs(bigcoeff, bigeval, bigevec, 2*N);

  det = 1.0;
  for(k=0; k<2*N; k++) det *= bigeval[k];
  if(det < 0.0) return -1000000000.0;
  if(!((bigcinv[0] > -1000000000.0) && (bigcinv[0] < 1000000000.0))) /* nan */
    return -1000000000.0;
  logl = -0.5*log(det);
  for(n=0; n<2*N; n++)
  {
    for(n2=0; n2<n; n2++) /* n2<n due to symmetry */
      logl -= bigdata[n]*bigdata[n2]*bigcinv[n+2*N*n2]; /* 2.0 * 0.5 = 1.0 */
    /* n2 = n; */
      logl -= 0.5*bigdata[n]*bigdata[n2]*bigcinv[n+2*N*n2];
  }
  if(VERBOSE) fprintf(stderr,"h2=%.03f %.03f %.03f %.03f  rho=%.03f xi2=%.03f %.03f => det=%e logl=%.03f\n",h2IFB,h2cisIFB,h2IFA,h2cisIFA,rho,xi2,xi2cis,det,logl); 
  if(!((logl > -1000000000.0) && (logl < 1000000000.0))) /* nan */
    return -1000000000.0;
  return logl;
}
