#include <stdio.h>
#include <nicklib.h>
#include <stdlib.h>
#include "eigsubs.h"
#include <unistd.h>

#define N 687 /* size of matrix */
#define P 2432 /* number of pairs */
#define MAXN 1500 /* max sample number */
#define M 18735 /* number of genes = 18735 */
#define INFILE "/groups/price/aprice/DECODEGE2/PHASED/FAMGEBIBD.782.norm"
#define THETAFILE "/groups/price/aprice/DECODEGE2/NEW/IFB.theta.txt"
#define PAIRSFILE "/groups/price/aprice/DECODEGE2/PHASED/FAMALLIBD"
#define ITER 15
#define VERBOSE 0

double *eval, *evec, *SUM, *data, h2, h2cis, *theta;
double *gammax;
double *coeff, *cinv, temp;
double compute_logl();
double compute_logl_h2cis();

int main(int argc, char **argv)
{
  double step, besth2, besth2cis, logl, bestlogl;
  int k, n, n2, m, tempn, i, iter, p, N1, N2;
  int *goodp, ngoodp, *nN, *Nn, *pn1, *pn2, mstart, mend;
  char tempstring[3000], Xchar;
  char *OUTFILE = NULL;
  FILE *fpin, *fptheta, *fppairs, *fpout;

  while((i = getopt(argc,argv,"s:e:o:")) != -1)
    {
      switch(i)
        {
        case 's':
          mstart = atoi(optarg);
          break;
        case 'e':
          mend = atof(optarg);
          break;
        case 'o':          /* input genotype file */
          OUTFILE = (char *) strdup(optarg);
          break;
        }
    }

  if((coeff = (double *) malloc(N*N*sizeof(*coeff))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((cinv = (double *) malloc(N*N*sizeof(*cinv))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((eval = (double *) malloc(N*sizeof(*eval))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((evec = (double *) malloc(N*N*sizeof(*evec))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((SUM = (double *) malloc(N*sizeof(*SUM))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((theta = (double *) malloc(N*N*sizeof(*theta))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((gammax = (double *) malloc(N*N*sizeof(*gammax))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((data = (double *) malloc(N*N*sizeof(*data))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((nN = (int *) malloc(N*sizeof(*nN))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((Nn = (int *) malloc(MAXN*sizeof(*Nn))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((goodp = (int *) malloc(P*sizeof(*goodp))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((pn1 = (int *) malloc(P*sizeof(*pn1))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  if((pn2 = (int *) malloc(P*sizeof(*pn2))) == NULL)
    { fprintf(stderr,"CM\n");  exit(1); }
  
  if( (fpin = fopen(INFILE, "r")) == NULL)
    {
      fprintf(stderr,"Could not open input file %s\n", INFILE);  exit(1);
    }
  if( (fptheta = fopen(THETAFILE, "r")) == NULL)
    {
      fprintf(stderr,"Could not open input file %s\n", THETAFILE);  exit(1);
    }
  if( (fppairs = fopen(PAIRSFILE, "r")) == NULL)
    {
      fprintf(stderr,"Could not open input file %s\n", THETAFILE);  exit(1);
    }
  if( (fpout = fopen(OUTFILE, "w")) == NULL)
    {
      fprintf(stderr,"Could not open output file %s\n", OUTFILE);  exit(1);
    }

  /* get theta from THETAFILE */
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<N; n2++) fscanf(fptheta,"%lf",&theta[n+N*n2]);
  }

  /* PROCESS 1ST LINE OF DATA FILE TO MAKE nN[] and Nn[] */
  fscanf(fpin,"%s",&tempstring); 
  for(k=0; k<MAXN; k++) Nn[k] = -1; /* -1 means absent */
  for(n=0; n<N; n++)
  {
    fscanf(fpin,"%d",&nN[n]);
    Nn[nN[n]] = n;
  }
  /* use PAIRSFILE to make pn1[p] and pn2[p] */
  ngoodp = 0;
  for(p=0; p<P; p++)
  {
    goodp[p] = 0;
    fscanf(fppairs,"%lf",&temp);
    fscanf(fppairs,"%d",&N1);
    fscanf(fppairs,"%d",&N2);
    if(Nn[N1]<0) continue; if(Nn[N2]<0) continue;
    goodp[p] = 1; ngoodp++;
    pn1[p] = Nn[N1]; pn2[p] = Nn[N2]; 
  }
  /* PROCESS GENE EXPRESSION DATA */
  for(m=0; m<M; m++)
  {
    /* make data[] and gammax[] */
    fscanf(fpin,"%s",&tempstring);
    for(n=0; n<N; n++) fscanf(fpin,"%lf",&data[n]);
    fscanf(fpin,"%c",&Xchar); /* space character */
    for(n=0; n<N; n++)
    { 
      for(n2=0; n2<N; n2++) gammax[n+N*n2] = 0.0;
    }
    for(p=0; p<P; p++)
    {
      fscanf(fpin,"%c",&Xchar);
      if(goodp[p] == 0) continue;
      if(Xchar == '1') 
        { gammax[pn1[p]+N*pn2[p]] = 0.5; gammax[pn2[p]+N*pn1[p]] = 0.5; }
      if(Xchar == '2') 
        { gammax[pn1[p]+N*pn2[p]] = 1.0; gammax[pn2[p]+N*pn1[p]] = 1.0; }
    }
    if(m < mstart) continue;
    if(m >= mend) break;

    /* ESTIMATE h2: binary search */
    h2 = 0.50; logl = compute_logl();
    besth2 = h2; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2 = besth2-step; logl = compute_logl();
      if(logl > bestlogl) { besth2=h2; bestlogl=logl; step /= 2.0; continue; }
      h2 = besth2+step; logl = compute_logl();
      if(logl > bestlogl) { besth2=h2; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2 = besth2;

    /* ESTIMATE h2cis: binary search */
    h2cis = 0.50; logl = compute_logl_h2cis();
    besth2cis = h2cis; bestlogl = logl; step = 0.5;
    /* binary search */
    for(iter=0; iter<ITER; iter++)
    {
      h2cis = besth2cis-step; logl = compute_logl_h2cis();
      if(logl > bestlogl) 
        { besth2cis=h2cis; bestlogl=logl; step /= 2.0; continue; }
      h2cis = besth2cis+step; logl = compute_logl_h2cis();
      if(logl > bestlogl) 
        { besth2cis=h2cis; bestlogl=logl; step /= 2.0; continue; }
      step /= 2.0;
    }
    h2cis = besth2cis;
    fprintf(fpout,"%.03f %.03f %.03f\n",h2,h2cis,h2-h2cis);
  }
}

double compute_logl()
{
  int k, n, n2;
  double det, logl;

  /* spectral decomposition */
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<N; n2++) 
      coeff[n+N*n2] = h2*theta[n+N*n2];
      /* abuse of notation: in this program, theta=0 on diagonal */
    coeff[n+N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  pdinv(cinv, coeff, N); 
  eigvecs(coeff, eval, evec, N);

  det = 1.0;
  for(k=0; k<N; k++) det *= eval[k];
  if(det < 0.0) return -1000000000.0;
  logl = -0.5*log(det);
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<n; n2++) /* n2<n due to symmetry */
      logl -= data[n]*data[n2]*cinv[n+N*n2]; /* 2.0 * 0.5 = 1.0 */
    n2 = n;
      logl -= 0.5*data[n]*data[n2]*cinv[n+N*n2];
  }
  if(VERBOSE) fprintf(stderr,"h2=%.03f h2cis=%.03f logl=%.03f\n",h2,h2cis,logl);
  if(!((logl > -1000000000.0) && (logl < 1000000000.0))) /* nan */
    return -1000000000.0;
  return logl;
}

double compute_logl_h2cis()
{
  int k, n, n2;
  double det, logl;

  /* spectral decomposition */
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<N; n2++) 
      coeff[n+N*n2] = h2cis*gammax[n+N*n2] + (h2-h2cis)*theta[n+N*n2];
      /* abuse of notation: in this program, theta=gamma=0 on diagonal */
    coeff[n+N*n] += 1.0; /* put diagonal back in variance matrix */
  }
  pdinv(cinv, coeff, N); 
  eigvecs(coeff, eval, evec, N);

  det = 1.0;
  for(k=0; k<N; k++) det *= eval[k];
  if(det < 0.0) return -1000000000.0;
  logl = -0.5*log(det);
  for(n=0; n<N; n++)
  {
    for(n2=0; n2<n; n2++) /* n2<n due to symmetry */
      logl -= data[n]*data[n2]*cinv[n+N*n2]; /* 2.0 * 0.5 = 1.0 */
    n2 = n;
      logl -= 0.5*data[n]*data[n2]*cinv[n+N*n2];
  }
  if(VERBOSE) fprintf(stderr,"h2=%.03f h2cis=%.03f logl=%.03f\n",h2,h2cis,logl);
  if(!((logl > -1000000000.0) && (logl < 1000000000.0))) /* nan */
    return -1000000000.0;
  return logl;
}
